"""
Scripts used in :mod:`.run_script` and :mod:`.setup_autopilot` to install packages and configure the system environment

Scripts are contained in the :data:`.scripts.SCRIPTS` dictionary, and each script is of the form::

    'script_name': {
        'type': 'bool', # always bool, signals that gui elements should present it as a checkbox to run or not
        'text': 'human readable description of what the script does',
        'commands': [
            'list of shell commands'
        ]
    }

The  commands in each ``commands`` list are concatenated with ``&&``  and run sequentially (see
:func:`.run_script.call_series` ). Certain commands that are expected to fail but don't impact the outcome of the
rest of the script -- eg. making a directory that already exists -- can be made optional by using the syntax::

    [
        'required command',
        {'command':'optional command', 'optional': True}
    ]

This concatenates the command with a ``; `` which doesn't raise an error if the command fails and allows the rest of
the script to proceed.

.. note::

    The above syntax will be used in the future for additional parameterizations that need to be made to scripts (
    though being optional is the only paramaterization avaialable now).

.. note::

    An unadvertised feature of ``raspi-config`` is the ability to run commands frmo the cli --
    find the name of a command here: https://github.com/RPi-Distro/raspi-config/blob/master/raspi-config
    and then use it like this: ``sudo raspi-config nonint function_name argument`` , so for example
    to enable the camera one just calls ``sudo raspi-config nonint do_camera 0`` (where turning the
    camera on, perhaps counterintuitively, is ``0`` which is true for all commands)

.. todo::

    Probably should have these use :data:`.prefs.get('S')copes` as well
"""

from collections import OrderedDict as odict

SCRIPTS = odict({
    'env_pilot': {
        'type': 'bool',
        'text': 'install system packages necessary for autopilot Pilots? (required if they arent already)',
        'commands': [
            "sudo apt-get update",
            "sudo apt-get install -y build-essential cmake git python3-dev libatlas-base-dev libsamplerate0-dev libsndfile1-dev libreadline-dev libasound-dev i2c-tools libportmidi-dev liblo-dev libhdf5-dev libzmq-dev libffi-dev",
        ]
    },
    'env_terminal': {
        'type': 'bool',
        'text': 'install system packages necessary for autopilot Terminals? (required if they arent already)',
        'commands': [
            'sudo apt-get update',
            'sudo apt-get install -y \
                libxcb-icccm4 \
                libxcb-image0 \
                libxcb-keysyms1 \
                libxcb-randr0 \
                libxcb-render-util0 \
                libxcb-xinerama0 \
                libxcb-xfixes0'
        ]
    },
    'performance': {
        'type': 'bool',
        'text': 'Do performance enhancements? (recommended, change cpu governor and give more memory to audio)',
        'commands': [
            'sudo systemctl disable raspi-config',
            'sudo sed -i \'/^exit 0/i echo "performance" | sudo tee /sys/devices/system/cpu/cpu*/cpufreq/scaling_governor\' /etc/rc.local',
            'sudo sh -c "echo @audio - memlock 256000 >> /etc/security/limits.conf"',
            'sudo sh -c "echo @audio - rtprio 75 >> /etc/security/limits.conf"',
            'sudo sh -c "echo vm.swappiness = 10 >> /etc/sysctl.conf"' # https://www.raspberrypi.org/forums/viewtopic.php?t=198765
        ]
    },
    'change_pw': {
        'type': 'bool',
        'text': "If you haven't, you should change the default raspberry pi password or you _will_ get your identity stolen. Change it now?",
        'commands': ['passwd']
    },
    'set_locale': {
        'type': 'bool',
        'text': 'Would you like to set your locale?',
        'commands': ['sudo dpkg-reconfigure locales',
                     'sudo dpkg-reconfigure keyboard-configuration']
    },
    'hifiberry': {
        'type': 'bool',
        'text': 'Setup Hifiberry DAC/AMP?',
        'commands': [
            {'command': 'sudo adduser pi i2c', 'optional': True},
            'sudo sed -i \'s/^dtparam=audio=on/#dtparam=audio=on/g\' /boot/config.txt',
            'sudo sed -i \'$s/$/\\ndtoverlay=hifiberry-dacplus\\ndtoverlay=i2s-mmap\\ndtoverlay=i2c-mmap\\ndtparam=i2c1=on\\ndtparam=i2c_arm=on/\' /boot/config.txt',
            'echo -e \'pcm.!default {\\n type hw card 0\\n}\\nctl.!default {\\n type hw card 0\\n}\' | sudo tee /etc/asound.conf'
        ]
    },
    # 'viz': {
    #     'type': 'bool',
    #     'text': 'Install X11 server and psychopy for visual stimuli?'
    # },
    'bluetooth': {
        'type': 'bool',
        'text': 'Disable Bluetooth? (recommended unless you\'re using it <3',
        'commands': [
            'sudo sed - i \'$s/$/\ndtoverlay=pi3-disable-bt/\' /boot/config.txt',
            'sudo systemctl disable hciuart.service',
            'sudo systemctl disable bluealsa.service',
            'sudo systemctl disable bluetooth.service'
        ],
    },
    'systemd': {
        'type': 'bool',
        'text': 'Install Autopilot as a systemd service?\nIf you are running this command in a virtual environment it will be used to launch Autopilot'
    },
    'alias': {
        'type': 'bool',
        'text': 'Create an alias to launch with "autopilot" (must be run from setup_autopilot, calls make_alias)'
    },
    'jackd_apt': {
        'type': 'bool',
        'text': 'Install jack audio from apt repository (required if AUDIOSERVER == jack)',
        'commands': ['sudo apt update && sudo apt install -y jackd2']
    },
    'jackd_source': {
        'type': 'bool',
        'text': 'Install jack audio from source, try this if youre having compatibility or runtime issues with jack (required if AUDIOSERVER == jack)',
        'commands': [
            "git clone git://github.com/jackaudio/jack2 --depth 1",
            "cd jack2",
            "./waf configure --alsa=yes --libdir=/usr/lib/arm-linux-gnueabihf/",
            "./waf build -j6",
            "sudo ./waf install",
            "sudo ldconfig",
            "sudo sh -c \"echo @audio - memlock 256000 >> /etc/security/limits.conf\"",  # giving jack more juice
            "sudo sh -c \"echo @audio - rtprio 75 >> /etc/security/limits.conf\"",
            "cd ..",
            "rm -rf ./jack2"
        ]
    },
    'opencv': {
        'type': 'bool',
        'text': 'Install OpenCV from source, including performance enhancements for ARM processors (takes awhile)',
        'commands': [
            "sudo apt-get install -y build-essential cmake ccache unzip pkg-config libjpeg-dev libpng-dev libtiff-dev libavcodec-dev libavformat-dev libswscale-dev libv4l-dev libxvidcore-dev libx264-dev ffmpeg libgtk-3-dev libcanberra-gtk* libatlas-base-dev gfortran python2-dev python-numpy",
            "git clone https://github.com/opencv/opencv.git",
            "git clone https://github.com/opencv/opencv_contrib",
            "cd opencv",
            "mkdir build",
            "cd build",
            "cmake -D CMAKE_BUILD_TYPE=RELEASE \
                -D CMAKE_INSTALL_PREFIX=/usr/local \
                -D OPENCV_EXTRA_MODULES_PATH=/home/pi/git/opencv_contrib/modules \
                -D BUILD_TESTS=OFF \
                -D BUILD_PERF_TESTS=OFF \
                -D BUILD_DOCS=OFF \
                -D WITH_TBB=ON \
                -D CMAKE_CXX_FLAGS=\"-DTBB_USE_GCC_BUILTINS=1 -D__TBB_64BIT_ATOMICS=0\" \
                -D WITH_OPENMP=ON \
                -D WITH_IPP=OFF \
                -D WITH_OPENCL=ON \
                -D WITH_V4L=ON \
                -D WITH_LIBV4L=ON \
                -D ENABLE_NEON=ON \
                -D ENABLE_VFPV3=ON \
                -D PYTHON3_EXECUTABLE=/usr/bin/python3 \
                -D PYTHON_INCLUDE_DIR=/usr/include/python3.7 \
                -D PYTHON_INCLUDE_DIR2=/usr/include/arm-linux-gnueabihf/python3.7 \
                -D OPENCV_ENABLE_NONFREE=ON \
                -D INSTALL_PYTHON_EXAMPLES=OFF \
                -D WITH_CAROTENE=ON \
                -D CMAKE_SHARED_LINKER_FLAGS='-latomic' \
                -D BUILD_EXAMPLES=OFF ..",
            "sudo sed -i 's/^CONF_SWAPSIZE=100/CONF_SWAPSIZE=2048/g' /etc/dphys-swapfile", # increase size of swapfile so multicore build works
            "sudo /etc/init.d/dphys-swapfile stop",
            "sudo /etc/init.d/dphys-swapfile start",
            "make -j4",
            "sudo --preserve-env=PATH make install",
            "sudo ldconfig",
            "sudo sed -i 's/^CONF_SWAPSIZE=2048/CONF_SWAPSIZE=100/g' /etc/dphys-swapfile",
            "sudo /etc/init.d/dphys-swapfile stop",
            "sudo /etc/init.d/dphys-swapfile start"
        ]
    },
    'performance_cameras': {
        'type': 'bool',
        'text': 'Do performance enhancements for video - mods to uvcvideo and increasing usbfs',
        'commands': [
            "sudo sh -c 'echo options uvcvideo nodrop=1 timeout=10000 quirks=0x80 > /etc/modprobe.d/uvcvideo.conf'",
            "sudo rmmod uvcvideo",
            "sudo modprobe uvcvideo",
            "sudo sed -i \"/^exit 0/i sudo sh -c 'echo ${usbfs_size} > /sys/module/usbcore/parameters/usbfs_memory_mb'\" /etc/rc.local"
        ],
    },
    'picamera': {
        'type': 'bool',
        'text': 'Enable PiCamera (with raspi-config)',
        'commands': [
            'sudo raspi-config nonint do_camera 0'
        ]
    },
    'pigpiod': {
        'type': 'bool',
        'text': 'Install pigpio daemon (sneakers fork that gives full timestamps and has greater capacity for scripts)',
        'commands': [
            'wget https://github.com/sneakers-the-rat/pigpio/archive/master.zip',
            'unzip master.zip',
            'cd pigpio-master',
            'make -j4',
            'sudo --preserve-env=PATH make install',
            'cd ..',
            'sudo rm -rf ./pigpio-master',
            'sudo rm ./master.zip'
        ]
    },
    'i2c': {
        'type': 'bool',
        'text': 'Enable i2c and set baudrate to 100kHz',
        'commands': [
            'sudo sed -i \'s/^#dtparam=i2c_arm=on/dtparam=i2c_arm=on/g\' /boot/config.txt',
            'sudo sed -i \'$s/$/\ni2c_arm_baudrate=100000/\' /boot/config.txt',
            'sudo sed -i \'$s/$/\ni2c-dev/\' /etc/modules',
            'sudo dtparam i2c_arm=on',
            'sudo modprobe i2c-dev'
        ]
    }
})
""""""



"""
Generic utility functions that are used in multiple places in the library that
for now don't have a clear other place to be
"""

import inspect
import importlib
import json
import sys
from pathlib import Path
import pkgutil
import ast
import typing
from threading import Thread
import numpy as np


def list_classes(module) -> typing.List[typing.Tuple[str, str]]:
    """
    List all classes within a module/package without importing by parsing the syntax tree
    directly with :mod:`ast` .

    Args:
        module (module, str): either the imported module to be queried, or its name as a string.
           if passed a string, attempt to import with :func:`importlib.import_module`

    Returns:
        list of tuples [('ClassName', 'module1.module2.ClassName')] a la :func:`inspect.getmembers`
    """
    ret_classes = []

    if not inspect.ismodule(module):
        module = importlib.import_module(module)

    # First get any members that are defined within the base __init__.py module itself
    base_classes = inspect.getmembers(module, inspect.isclass)
    ret_classes.extend([
        (bc[0], ".".join([bc[1].__module__, bc[1].__name__]))
        for bc in base_classes
    ])

    # get the parent directory name and module name, we'll use that
    mod_path = Path(module.__file__).resolve().parent
    if '__init__' not in str(Path(module.__file__)):
        mod_name = ".".join(module.__name__.split('.')[:-1])
    else:
        mod_name = module.__name__

    # get names of module files within top-level package
    submodules = [mod for _, mod, _ in pkgutil.iter_modules([mod_path])]
    submod_paths = [(mod_path / mod).with_suffix('.py') for mod in submodules]

    # parse the files to get the names of the classes
    for submod_name, submod in zip(submodules, submod_paths):
        with open(submod, 'r') as submod_f:
            submod_ast = ast.parse(submod_f.read())

        ret_classes.extend([
            (n.name, '.'.join([mod_name, submod_name, n.name]))
            for n in submod_ast.body
            if isinstance(n, ast.ClassDef)
        ])

    return ret_classes


def find_class(cls_str:str):
    """
    Given a full package.module.ClassName string, return the relevant class

    Args:
        cls_str (str): a full package.module.ClassName string, like ``'autopilot.hardware.Hardware'``

    Returns:
        the class indicated by ``cls_str``
    """
    # get class and module names first by splitting off final class name
    class_name = cls_str.split('.')[-1]
    module_name = '.'.join(cls_str.split('.')[:-1])

    # import or get reference to already-imported module
    if module_name in sys.modules:
        mod = sys.modules[module_name]
    else:
        mod = importlib.import_module(module_name)

    return getattr(mod, class_name)


def recurse_subclasses(cls, leaves_only=False) -> list:
    """
    Given some class, find its subclasses recursively

    See: https://stackoverflow.com/a/17246726/13113166

    Args:
        leaves_only (bool): If True, only include classes that have no further subclasses,
            if False (default), return all subclasses.

    Returns:
        list of subclasses
    """

    all_subclasses = []

    for subclass in cls.__subclasses__():
        if leaves_only:
            if len(subclass.__subclasses__()) == 0:
                all_subclasses.append(subclass)
        else:
            all_subclasses.append(subclass)
        all_subclasses.extend(recurse_subclasses(subclass, leaves_only))

    return all_subclasses


class ReturnThread(Thread):
    """
    Thread whose .join() method returns the value from the function
    thx to https://stackoverflow.com/a/6894023
    """
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs, Verbose)
        self._return = None
    def run(self):
        if self._Thread__target is not None:
            self._return = self._Thread__target(*self._Thread__args,
                                                **self._Thread__kwargs)
    def join(self, timeout=None):
        Thread.join(self, timeout)

        return self._return


def list_subjects(pilot_db=None):
    """
    Given a dictionary of a pilot_db, return the subjects that are in it.

    Args:
        pilot_db (dict): a pilot_db. if None tried to load pilot_db with :method:`.load_pilotdb`

    Returns:
        subjects (list): a list of currently active subjects

    """

    if pilot_db is None:
        pilot_db = load_pilotdb()

    subjects = []
    for pilot, values in pilot_db.items():
        if 'subjects' in values.keys():
            subjects.extend(values['subjects'])

    return subjects


def load_pilotdb(file_name=None, reverse=False):
    """
    Try to load the file_db

    Args:
        reverse:
        file_name:

    Returns:

    """

    if file_name is None:
        file_name = '/usr/autopilot/pilot_db.json'

    with open(file_name) as pilot_file:
        pilot_db = json.load(pilot_file)

    if reverse:
        # simplify pilot db
        pilot_db = {k: v['subjects'] for k, v in pilot_db.items()}
        pilot_dict = {}
        for pilot, subjectlist in pilot_db.items():
            for ms in subjectlist:
                pilot_dict[ms] = pilot
        pilot_db = pilot_dict

    return pilot_db


def coerce_discrete(df, col, mapping={'L':0, 'R':1}):
    """
    Coerce a discrete/string column of a pandas dataframe into numeric values

    Default is to map 'L' to 0 and 'R' to 1 as in the case of Left/Right 2AFC tasks

    Args:
        df (:class:`pandas.DataFrame`) : dataframe with the column to transform
        col (str):  name of column
        mapping (dict): mapping of strings to numbers

    Returns:
        df (:class:`pandas.DataFrame`) : transformed dataframe

    """

    for key, val in mapping.items():
        df.loc[df[col]==key,col] = val

    # if blanks, warn and remove
    if '' in df[col].unique():
        n_blanks = sum(df[col]=='')
        Warning('{} blank rows detected, removing.'.format(n_blanks))
        df.drop(df.index[df[col]==''], axis=0, inplace=True)

    df = df.astype({col:float})
    return df


def find_key_recursive(key, dictionary):
    """
    Find all instances of a key in a dictionary, recursively.

    Args:
        key:
        dictionary:

    Returns:
        list
    """
    for k, v in dictionary.items():
        if k == key:
            yield v
        elif isinstance(v, dict):
            for result in find_key_recursive(key, v):
                yield result
        elif isinstance(v, list):
            for d in v:
                for result in find_key_recursive(key, d):
                    yield result

def find_key_value(dicts:typing.List[dict], key:str, value:str, single=True):
    """
    Find an entry in a list of dictionaries where dict[key] == value.

    Args:
        dicts ():
        key ():
        value ():
        single (bool): if ``True`` (default), raise an exception if multiple
            results are matched
    """
    if not all([isinstance(item, dict) for item in dicts]):
        raise ValueError(f"Pass me a list of dicts! got {dicts}")

    matches = [match for match in dicts if match[key] == value]

    if len(matches)>1 and single:
        raise IndexError(f"Multiple matches found for key: {key}, value: {value}, got:\n{matches}")
    elif len(matches) == 1:
        matches = matches[0]

    return matches


class NumpyEncoder(json.JSONEncoder):
    """
    Allow json serialization of objects containing numpy arrays.

    Use like ``json.dump(obj, fp, cls=NumpyEncoder)``

    Deserialize with :class:`.NumpyDecoder`

    References:
        * https://stackoverflow.com/a/49677241/13113166
        * https://github.com/mpld3/mpld3/issues/434#issuecomment-340255689
        * https://gist.github.com/massgh/297a73f2dba017ffd28dbc34b9a40e90
    """

    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):  #### This is the fix
            return {
                "_kind_": "ndarray",
                "_value_": obj.tolist()
            }
        return json.JSONEncoder.default(self, obj)

class NumpyDecoder(json.JSONDecoder):
    """
    Allow json deserialization of objects containing numpy arrays.

    Use like ``json.load(fp, cls=NumpyDecoder)``

    Serialize with :class:`.NumpyEncoder`

    References:
        * https://stackoverflow.com/a/49677241/13113166
        * https://github.com/mpld3/mpld3/issues/434#issuecomment-340255689
        * https://gist.github.com/massgh/297a73f2dba017ffd28dbc34b9a40e90
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        try:
            kind = obj['_kind_']
            if kind == 'ndarray':
                    obj = np.array(obj['_value_'])

        except (TypeError, KeyError):
            # normal, just return obj
            pass

        return obj


import typing
import warnings

import numpy as np

from autopilot.transform.transforms import Transform
from autopilot.transform.image import DLC


class Slice(Transform):
    """
    Generic selection processor
    """
    format_in = {'type': 'any'}
    format_out = {'type': 'any'}

    def __init__(self, select, *args, **kwargs):
        """


        Args:
            select (slice, tuple[slice], int, tuple[int]): a slice, tuple of slices, int, or tuple of ints! anything you can
                use inside of a pair of [square brackets].
            *args:
            **kwargs:
        """
        super(Slice, self).__init__(*args, **kwargs)

        # self.check_slice(select)

        self.select = select

    # def check_slice(self, select):
    #     if isinstance(select, tuple):
    #         if not all([isinstance(inner, slice) for inner in select]):
    #             raise ValueError('Selections require slices or tuples of slices')
    #     elif not isinstance(select, slice):
    #         raise ValueError('Selections require slices or tuples of slices')



    def process(self, input):
        return input[self.select]


class DLCSlice(Slice):
    """
    Select x,y coordinates of :class:`.DLC` output based on the name of the tracked parts

    note that min_probability is undefined when a list or tuple of part names are defined:
    the form of the returned array is ambiguous (how to tell which part is which when some might be excluded?)

    """
    format_in = {'type': np.ndarray,
                 'parent': DLC}
    format_out = {'type': np.ndarray}

    def __init__(self, select: typing.Union[str, tuple, list],
                 min_probability: float = 0,  *args, **kwargs):
        super(DLCSlice, self).__init__(select, *args, **kwargs)

        self.select_index = None

        if isinstance(select, (tuple, list)) and min_probability > 0:
            warnings.warn('min_probability is undefined when a list or tuple of part names are given, ignoring.')

        self.min_probability = np.clip(min_probability, 0, 1)

    def check_slice(self, select):
        if self._parent:
            # only check if we've already gotten a parent
            if isinstance(select, str):
                if select not in self._parent.live.cfg['all_joints_names']:
                    raise ValueError('DLC selections must be names of joints!')
            elif isinstance(select, (tuple, list)):
                for s in select:
                    if s not in self._parent.live.cfg['all_joints_names']:
                        raise ValueError('DLC selections must be names of joints!')

    def process(self, input: np.ndarray):
        if self.select_index is None:
            if isinstance(self.select, str):
                self.select_index = self._parent.live.cfg['all_joints_names'].index(self.select)
            else:
                self.select_index = np.array([self._parent.live.cfg['all_joints_names'].index(s) for s in self.select])

        point_row = input[self.select_index, :]
        if isinstance(self.select, str):
            if point_row[2] > self.min_probability:
                return point_row[0:2]
            else:
                return False
        else:
            return point_row